(* Developped by DooMeeR - webmaster@doomeer.com *)
unit Partition;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

type
  TPartitionNote = class
    Note: byte;
    Len: integer;
    Pos: integer;
    Tag: integer;
    Sel: boolean;
  end;
  TClickRec = record
    Pos: integer;
    LargePos: integer;
    Note: TPartitionNote;
  end;
  pClickRec = ^TClickRec;
  TExtMouseEvent = procedure(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; Pos, LargePos: integer; NoteNum: byte; Note: TPartitionNote) of object;
  TExtMouseMoveEvent = procedure(Sender: TObject; Shift: TShiftState; X, Y: Integer; Pos, LargePos: integer; NoteNum: byte; Note: TPartitionNote) of object;
  TPartition = class(TGraphicControl)
  private
    FInterWidth: byte;
    FSmallestWidth: byte;
    FLargePosLen: byte;
    FOnExtMouseDown: TExtMouseEvent;
    FOnExtMouseMove: TExtMouseMoveEvent;
    FOnExtMouseUp: TExtMouseEvent;
    FCursorPos: integer;
    FMesureLen: integer;
    FVertGridLen: integer;
    FFirstMesure: integer;
    FFirstPixel: integer;
    FFirstPos: integer;
    procedure SetInterWidth(const Value: byte);
    procedure SetSmallestWidth(const Value: byte);
    procedure SetCursorPos(const Value: integer);
    procedure SetLargePosLen(const Value: byte);
    procedure SetMesureLen(const Value: integer);
    procedure SetVertGridLen(const Value: integer);
    procedure SetFirstMesure(const Value: integer);
  protected
    LastClick: pClickRec;
    PaintSave: TBitmap;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure ExtMouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer; Pos, LargePos: integer; NoteNum: byte; Note: TPartitionNote); virtual;
    procedure ExtMouseMove(Shift: TShiftState; X, Y: Integer; Pos, LargePos: integer; NoteNum: byte; Note: TPartitionNote); virtual;
    procedure ExtMouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer; Pos, LargePos: integer; NoteNum: byte; Note: TPartitionNote); virtual;
  public
    Notes: TList;
    CanPaint: boolean;
    procedure Paint; override;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddNote(Note: byte; Pos, Len, Tag: integer);
    procedure DeleteNote(Note: TPartitionNote);
    function GetFirstNoteAt(Pos: integer; Note: byte): TPartitionNote;
    function GetNote(Y: integer): byte;
    procedure Clear;
    procedure ClearSel;
    property Canvas;
  published
    property InterWidth: byte read FInterWidth write SetInterWidth;
    property SmallestWidth: byte read FSmallestWidth write SetSmallestWidth;
    property LargePosLen: byte read FLargePosLen write SetLargePosLen;
    property CursorPos: integer read FCursorPos write SetCursorPos;
    property MesureLen: integer read FMesureLen write SetMesureLen;
    property VertGridLen: integer read FVertGridLen write SetVertGridLen;
    property FirstMesure: integer read FFirstMesure write SetFirstMesure;
    property FirstPixel: integer read FFirstPixel;
    property FirstPos: integer read FFirstPos;
    property Anchors;
    property Align;
    property OnClick;
    property OnDblClick;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnStartDrag;
    property OnExtMouseDown: TExtMouseEvent read FOnExtMouseDown write FOnExtMouseDown;
    property OnExtMouseMove: TExtMouseMoveEvent read FOnExtMouseMove write FOnExtMouseMove;
    property OnExtMouseUp: TExtMouseEvent read FOnExtMouseUp write FOnExtMouseUp;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Perso', [TPartition]);
end;

{ TPartition }

procedure TPartition.AddNote(Note: byte; Pos, Len, Tag: integer);
var N: TPartitionNote;
begin
  N:=TPartitionNote.Create;
  N.Note:=Note;
  N.Len:=Len;
  N.Pos:=Pos;
  N.Tag:=Tag;
  N.Sel:=false;
  Notes.Add(N);
  Paint;
end;

procedure TPartition.Clear;
var i: integer;
begin
  for i:=0 to Notes.Count-1 do TPartitionNote(Notes[i]).Free;
  Notes.Clear;
  Paint;
end;

procedure TPartition.ClearSel;
var i: integer;
begin
  for i:=0 to Notes.Count-1 do TPartitionNote(Notes[i]).Sel:=false;
  Paint;
end;

constructor TPartition.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Notes:=TList.Create;
  FInterWidth:=8;
  FSmallestWidth:=2;
  FLargePosLen:=4;
  FCursorPos:=-1;
  FMesureLen:=64;
  FVertGridLen:=16;
  FFirstMesure:=0;
  FFirstPixel:=0;
  FFirstPos:=0;
  Height:=112;
  Width:=300;
  CanPaint:=true;
  PaintSave:=TBitmap.Create;
{  if csDesigning in ComponentState then
  begin
  end; }
end;

procedure TPartition.DeleteNote(Note: TPartitionNote);
begin
  if Note<>nil then
    if Notes.IndexOf(Note)>=0 then
      Notes.Delete(Notes.IndexOf(Note));
  Paint;
end;

destructor TPartition.Destroy;
var i: integer;
    N: TPartitionNote;
begin
  for i:=0 to Notes.Count-1 do
  begin
    N:=Notes[i];
    N.Free;
  end;
  Notes.Free;
  PaintSave.Free;
  inherited Destroy;
end;

procedure TPartition.ExtMouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y, Pos, LargePos: integer; NoteNum: byte; Note: TPartitionNote);
begin
  if Assigned(FOnExtMouseDown) then FOnExtMouseDown(Self, Button, Shift, X, Y, Pos, LargePos, NoteNum, Note);
end;

procedure TPartition.ExtMouseMove(Shift: TShiftState; X, Y, Pos,
  LargePos: integer; NoteNum: byte; Note: TPartitionNote);
begin
  if Assigned(FOnExtMouseMove) then FOnExtMouseMove(Self, Shift, X, Y, Pos, LargePos, NoteNum, Note);
end;

procedure TPartition.ExtMouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y, Pos, LargePos: integer; NoteNum: byte; Note: TPartitionNote);
begin
  if Assigned(FOnExtMouseUp) then FOnExtMouseUp(Self, Button, Shift, X, Y, Pos, LargePos, NoteNum, Note);
end;

function TPartition.GetFirstNoteAt(Pos: integer; Note: byte): TPartitionNote;
var i: integer;
    N: TPartitionNote;
begin
  result:=nil;
  for i:=0 to Notes.Count-1 do
  begin
    N:=Notes[i];
    if N.Note = Note then
      if (N.Pos <= Pos) and (N.Pos+N.Len >= Pos) then
      begin
        result:=N;
        break;
      end;
  end;
  if result=nil then
  for i:=0 to Notes.Count-1 do
  begin
    N:=Notes[i];
    if N.Note = Note+1 then
      if (N.Pos <= Pos) and (N.Pos+N.Len >= Pos) then
      begin
        result:=N;
        break;
      end;
  end;
end;

function TPartition.GetNote(Y: integer): byte;
begin
  if (Y<=InterWidth*13.5) and (Y>InterWidth*13) then result:=36 else
  if (Y<=InterWidth*13) and (Y>InterWidth*12.5) then result:=38 else
  if (Y<=InterWidth*12.5) and (Y>InterWidth*12) then result:=40 else
  if (Y<=InterWidth*12) and (Y>InterWidth*11.5) then result:=41 else
  if (Y<=InterWidth*11.5) and (Y>InterWidth*11) then result:=43 else
  if (Y<=InterWidth*11) and (Y>InterWidth*10.5) then result:=45 else
  if (Y<=InterWidth*10.5) and (Y>InterWidth*10) then result:=47 else
  if (Y<=InterWidth*10) and (Y>InterWidth*9.5) then result:=48 else
  if (Y<=InterWidth*9.5) and (Y>InterWidth*9) then result:=50 else
  if (Y<=InterWidth*9) and (Y>InterWidth*8.5) then result:=52 else
  if (Y<=InterWidth*8.5) and (Y>InterWidth*8) then result:=53 else
  if (Y<=InterWidth*8) and (Y>InterWidth*7.5) then result:=55 else
  if (Y<=InterWidth*7.5) and (Y>InterWidth*7) then result:=57 else
  if (Y<=InterWidth*7) and (Y>InterWidth*6.5) then result:=59 else
  if (Y<=InterWidth*6.5) and (Y>InterWidth*6) then result:=60 else
  if (Y<=InterWidth*6) and (Y>InterWidth*5.5) then result:=62 else
  if (Y<=InterWidth*5.5) and (Y>InterWidth*5) then result:=64 else
  if (Y<=InterWidth*5) and (Y>InterWidth*4.5) then result:=65 else
  if (Y<=InterWidth*4.5) and (Y>InterWidth*4) then result:=67 else
  if (Y<=InterWidth*4) and (Y>InterWidth*3.5) then result:=69 else
  if (Y<=InterWidth*3.5) and (Y>InterWidth*3) then result:=71 else
  if (Y<=InterWidth*3) and (Y>InterWidth*2.5) then result:=72 else
  if (Y<=InterWidth*2.5) and (Y>InterWidth*2) then result:=74 else
  if (Y<=InterWidth*2) and (Y>InterWidth*1.5) then result:=76 else
  if (Y<=InterWidth*1.5) and (Y>InterWidth*1) then result:=77 else
  if (Y<=InterWidth*1) and (Y>InterWidth*0.5) then result:=79 else
  if (Y<=InterWidth*0.5) and (Y>InterWidth*0) then result:=81 else
    result:=0;
end;

procedure TPartition.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
var NoteNum: byte;
begin
  inherited MouseDown(Button, Shift, X, Y);
  if LastClick = nil then New(LastClick);
  LastClick.Pos:=(X div SmallestWidth)+FirstPos;
  LastClick.LargePos:=((X div (SmallestWidth * LargePosLen)) * LargePosLen)+FirstPos;
  NoteNum:=GetNote(Y);
  LastClick.Note:=GetFirstNoteAt(LastClick.Pos, NoteNum);
  ExtMouseDown(Button, Shift, X, Y, LastClick.Pos, LastClick.LargePos, NoteNum, LastClick.Note);
end;

procedure TPartition.MouseMove(Shift: TShiftState; X, Y: Integer);
var Pos, LargePos: integer;
    NoteNum: byte;
    Note: TPartitionNote;
begin
  inherited MouseMove(Shift, X, Y);
  Pos:=(X div SmallestWidth)+FirstPos;
  LargePos:=((X div (SmallestWidth * LargePosLen)) * LargePosLen)+FirstPos;
  NoteNum:=GetNote(Y);
  Note:=GetFirstNoteAt(Pos, NoteNum);
  ExtMouseMove(Shift, X, Y, Pos, LargePos, NoteNum, Note);
end;

procedure TPartition.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
var NewClick: TClickRec;
    NoteNum: byte;
begin
  inherited MouseUp(Button, Shift, X, Y);
  NewClick.Pos:=(X div SmallestWidth)+FirstPos;
  NewClick.LargePos:=((X div (SmallestWidth * LargePosLen)) * LargePosLen)+FirstPos;
  NoteNum:=GetNote(Y);
  NewClick.Note:=GetFirstNoteAt(NewClick.Pos, NoteNum);
  ExtMouseUp(Button, Shift, X, Y, NewClick.Pos, NewClick.LargePos, NoteNum, NewClick.Note);
  if LastClick <> nil then
  begin
    Dispose(LastClick);
    LastClick:=nil;
  end;
end;

procedure TPartition.Paint;
var i, p: integer;
    N: TPartitionNote;
begin
  inherited Paint;
  if not CanPaint then exit;
  PaintSave.Width:=Width;
  PaintSave.Height:=Height;
  with PaintSave.Canvas do
  begin
    // Fond
    Brush.Color:=clWhite;
    Brush.Style:=bsSolid;
    FillRect(Rect(0, 0, Width, Height));

    // Porte
    Pen.Color:=clBlack;
    Pen.Style:=psSolid;
    Pen.Width:=1;
    MoveTo(0, InterWidth*5);
    LineTo(Width, InterWidth*5);
    MoveTo(0, InterWidth*6);
    LineTo(Width, InterWidth*6);
    MoveTo(0, InterWidth*7);
    LineTo(Width, InterWidth*7);
    MoveTo(0, InterWidth*8);
    LineTo(Width, InterWidth*8);
    MoveTo(0, InterWidth*9);
    LineTo(Width, InterWidth*9);

    // Grille verticale
    for i:=0 to Width div (VertGridLen * SmallestWidth) do
    begin
      Pen.Color:=$B0B0B0;
      MoveTo(i * VertGridLen * SmallestWidth, 0);
      LineTo(i * VertGridLen * SmallestWidth, Height);
    end;

    // Mesures
    for i:=0 to Width div (MesureLen * SmallestWidth) do
    begin
      Pen.Color:=$B0B0B0;
      MoveTo(i * MesureLen * SmallestWidth, 0);
      LineTo(i * MesureLen * SmallestWidth, Height);
      Pen.Color:=clBlack;
      MoveTo(i * MesureLen * SmallestWidth, InterWidth*5);
      LineTo(i * MesureLen * SmallestWidth, InterWidth*9);
      Font.Color:=clBlack;
      Font.Name:='fixedsys';
      TextOut(i * MesureLen * SmallestWidth+2, 0, inttostr(i+FirstMesure));
    end;

    // Notes visibles
    for i:=0 to Notes.Count-1 do
    begin
      N:=Notes[i];
      if ((N.Pos - FirstPos)<=Width div SmallestWidth) and
         ((N.Pos+N.Len - FirstPos)>=0) then
      begin
        case N.Note of
          36, 37: p:=InterWidth*13;
          38, 39: p:=trunc(InterWidth*12.5);
          40: p:=InterWidth*12;
          41, 42: p:=trunc(InterWidth*11.5);
          43, 44: p:=InterWidth*11;
          45, 46: p:=trunc(InterWidth*10.5);
          47: p:=InterWidth*10;
          48, 49: p:=trunc(InterWidth*9.5);
          50, 51: p:=InterWidth*9;
          52: p:=trunc(InterWidth*8.5);
          53, 54: p:=InterWidth*8;
          55, 56: p:=trunc(InterWidth*7.5);
          57, 58: p:=InterWidth*7;
          59: p:=trunc(InterWidth*6.5);
          60, 61: p:=InterWidth*6;
          62, 63: p:=trunc(InterWidth*5.5);
          64: p:=InterWidth*5;
          65, 66: p:=trunc(InterWidth*4.5);
          67, 68: p:=InterWidth*4;
          69, 70: p:=trunc(InterWidth*3.5);
          71: p:=InterWidth*3;
          72, 73: p:=trunc(InterWidth*2.5);
          74, 75: p:=InterWidth*2;
          76: p:=trunc(InterWidth*1.5);
          77, 78: p:=InterWidth*1;
          79, 80: p:=trunc(InterWidth*0.5);
          81, 82: p:=0;
        else p:=-1;
        end;
        if p>=0 then
        begin
          case N.Note of
            36, 38, 40, 41, 43, 45, 47, 48, 50, 52, 53, 55, 57, 59, 60, 62, 64, 65, 67, 69, 71,
            72, 74, 76, 77, 79, 81: Brush.Color:=clWhite;
          else Brush.Color:=clYellow;
          end;
          if N.Sel then Brush.Color:=clBlack;
          Rectangle((N.Pos - FirstPos) * SmallestWidth, p+2, (N.Pos + N.Len - FirstPos) * SmallestWidth + 1, p + InterWidth - 1);
        end;
      end;
    end;
  end;
  // Dessin de la porte et des notes (calcul au dessus)
  Canvas.Draw(0, 0, PaintSave);
  // Curseur
  Canvas.Pen.Color:=clRed;
  Canvas.Pen.Style:=psSolid;
  Canvas.Pen.Width:=1;
  Canvas.MoveTo(CursorPos - FirstPixel, 0);
  Canvas.LineTo(CursorPos - FirstPixel, Height);
end;

procedure TPartition.SetCursorPos(const Value: integer);
var R: TRect;
begin
  if not (((Value - FirstPixel)>=0) and ((Value - FirstPixel)<=(Width - Width mod (MesureLen*SmallestWidth)))) then
    FirstMesure:=Value div (MesureLen*SmallestWidth);
  R:=Rect((FCursorPos - FirstPixel)-1, 0, (FCursorPos - FirstPixel)+1, Height);
  Canvas.CopyRect(R, PaintSave.Canvas, R);
  Canvas.Pen.Color:=clRed;
  Canvas.Pen.Style:=psSolid;
  Canvas.Pen.Width:=1;
  Canvas.MoveTo(Value - FirstPixel, 0);
  Canvas.LineTo(Value - FirstPixel, Height);
  FCursorPos:=Value;
end;

procedure TPartition.SetFirstMesure(const Value: integer);
begin
  FFirstMesure := Value;
  if FFirstMesure<0 then FFirstMesure:=0;
  FFirstPos := FFirstMesure * MesureLen;
  FFirstPixel := FFirstPos * SmallestWidth;
  Paint;
end;

procedure TPartition.SetInterWidth(const Value: byte);
begin
  FInterWidth:=Value;
  if FInterWidth<1 then FInterWidth:=1;
  Height:=Value*14;
  Refresh;
end;

procedure TPartition.SetLargePosLen(const Value: byte);
begin
  FLargePosLen := Value;
  if FLargePosLen<1 then FLargePosLen:=1;
end;

procedure TPartition.SetMesureLen(const Value: integer);
begin
  FMesureLen := Value;
  Paint;
end;

procedure TPartition.SetSmallestWidth(const Value: byte);
begin
  FSmallestWidth := Value;
  if FSmallestWidth<1 then FSmallestWidth:=1;
  Paint;
end;

procedure TPartition.SetVertGridLen(const Value: integer);
begin
  FVertGridLen := Value;
  Paint;
end;

end.
